/*
 * Copyright (C) 2021, Xilinx Inc - All rights reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

#include <memory>
#include <cstring>
#include <string.h>
#include <unistd.h>
#include "InternalEncoderSession.h"
#include "InternalFrame.h"
#include "InternalDataBuffer.h"
#include "xma.h"
#include "vpi_pic.h"
#include "vpi_video_vce_enc.h"

#ifndef MAX
#define MAX(a, b) (((a) > (b)) ? (a) : (b))
#endif

typedef struct VpiEncContext {
  VpiEncType vpi_encoder_type;
  void*      vpi_encoder_handle;
} VpiEncContext;

typedef struct _Xav1Context {
  VpiEncInitParams props;
  VpiEncHandle     xav1enc;
  int              poc;
  int              max_frames_delay;
  bool             isInputPoolExtended;
  void*            extradata;
} Xav1Context;

struct EnumName {
  int32_t     enum_value;
  const char* enum_name;
};

EnumName PresetList[] = {
    {XMA_ENC_PRESET_MEDIUM, "medium"},
    {XMA_ENC_PRESET_SLOW, "slow"},
};

EnumName H264ProfileList[] = {
    {XMA_ENC_PROFILE_AUTO, "auto"},
    {XMA_ENC_PROFILE_H264_BASELINE, "baseline"},
    {XMA_ENC_PROFILE_H264_MAIN, "main"},
    {XMA_ENC_PROFILE_H264_HIGH, "high"},
    {XMA_ENC_PROFILE_H264_HIGH10, "high-10"},
    {XMA_ENC_PROFILE_H264_HIGH10_INTRA, "high-10-intra"},
};

EnumName HEVCProfileList[] = {
    {XMA_ENC_PROFILE_AUTO, "auto"},
    {XMA_ENC_PROFILE_HEVC_MAIN, "main"},
    {XMA_ENC_PROFILE_HEVC_MAIN_INTRA, "main-intra"},
    {XMA_ENC_PROFILE_HEVC_MAIN10, "main-10"},
    {XMA_ENC_PROFILE_HEVC_MAIN10_INTRA, "main-10-intra"},
};

EnumName AV1ProfileList[] = {
    {XMA_ENC_PROFILE_AUTO, "auto"},
    {XMA_ENC_PROFILE_AV1_MAIN, "main"},
};

const char* GetEnumName(EnumName arr[], int32_t count, int32_t value) {
  for (int32_t index = 0; index < count; ++index) {
    if (arr[index].enum_value == value) {
      return arr[index].enum_name;
    }
  }
  return NULL;
}

namespace xma {

#define XMA_ENCODER_SESSION "EncoderSession"

///////////////////////////////////////////////////////////////////////////////
struct EncoderFrame {
public:
  EncoderFrame(FramePtr frame, VpiFrm* frm) : m_frame(frame), m_frm(frm) {
  }
  ~EncoderFrame() { vpi_frame_unref(m_frm); }

public:
  FramePtr m_frame;
  VpiFrm*  m_frm;
};

///////////////////////////////////////////////////////////////////////////////
InternalEncoderSession::InternalEncoderSession(IntSessionFactoryPtr sf, XmaEncoderProperties* props) : SessionBase(sf) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InternalEncoderSession()");

  InitParams(props);

  if (SetParam(props->params, props->param_cnt, XMA_ENC_PARAM_DELAY_INITIALIZATION, 0, 1, XMA_ENC_DELAY_INITIALIZATION_DEFAULT) == 0) {
    StartEncoder(nullptr);
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InternalEncoderSession() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
InternalEncoderSession::~InternalEncoderSession() {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::~InternalEncoderSession()");

  m_handle.reset();
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalEncoderSession::SetLog(LogPtr log) {
  lock_t guard(m_mutex);
  SessionBase::SetLog(log);
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalEncoderSession::Enqueue(queue_t& data) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Enqueue()");
  FramePtr frame = std::move(std::get<QueueFrame>(data).frame);

  if (!m_handle.get()) {
    StartEncoder(frame);
  }

  if (!frame) {
    lock_t guard(m_mutex);
    if (!m_end_found) {
      m_end_found = true;
      int ret;
      if ((ret = vpi_encoder_put_frame(m_handle.get(), nullptr)) != 0) {
        THROW_RUNTIME_EXCEPTION(
            m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Enqueue() Failed to send end of file to encoder (%s)!", vpi_error_str(ret));
      }
    }
    ++m_frames_processing;

    m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Enqueue() - Returning success");
    return XMA_SUCCESS;
  }

  VpiFrm* pic = nullptr;
  if ((vpi_frame_alloc(m_sf->GetMemHandle(), &pic) < 0) || !pic) {
    THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Enqueue() Out of memory!");
  }
  InitVpiFrm(pic);

  if (AddFrameToVpiFrm(m_sf, m_log, frame, pic) != XMA_SUCCESS) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Enqueue() Failed to add XmaFrame to VpiFrm!");
  }

  pic->poc = 0;
  pic->pts = frame->GetPts();
  pic->dts = 0;

  lock_t guard(m_mutex);

  if (vpi_encoder_put_frame(m_handle.get(), pic) < 0) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Enqueue() Failed to send frame to encoder!");
  }
  ++m_frames_processing;

  m_frame_sent++;
  if (m_latency_logging) {
    clock_gettime(CLOCK_REALTIME, &m_latency);
    m_time_taken = (m_latency.tv_sec * 1e3) + (m_latency.tv_nsec / 1e6);
    m_log->LogMsg(
        XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Enqueue() : %p : xma_enc_frame_sent : %lld : %lld\n", m_handle.get(), m_frame_sent, m_time_taken);
  }

  int32_t          ret    = XMA_SUCCESS;
  IntDataBufferPtr packet = PullPacket(&ret);
  if (ret == XMA_EOS) {
    m_queue.push_back(nullptr);
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Enqueue() - Returning success");
    return XMA_SUCCESS;
  } else if (ret == XMA_RESEND_AND_RECV) {
    if (m_frames_processing >= m_max_frames_processing) {
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Enqueue() - Returning success");
      return XMA_SUCCESS;
    } else {
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Enqueue() - Returning send more data");
      return XMA_SEND_MORE_DATA;
    }
  } else {
    m_queue.push_back(std::move(packet));
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Enqueue() - Returning success");
    return XMA_SUCCESS;
  }
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalEncoderSession::Dequeue(queue_t& data) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Dequeue()");

  if (!m_handle.get()) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Dequeue() Encoder not initialized!");
  }

  std::unique_lock<std::mutex> lock(m_mutex);

  if (!m_queue.empty()) {
    --m_frames_processing;
    if (m_queue.front() == nullptr) {
      m_queue.pop_front();
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Dequeue() - Returning EOS");
      return XMA_EOS;
    }
    data                                 = QueueDataBuffer();
    std::get<QueueDataBuffer>(data).data = std::move(m_queue.front());
    m_queue.pop_front();
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Dequeue() - Returning success");
    return XMA_SUCCESS;
  }

  int              ret;
  IntDataBufferPtr packet = PullPacket(&ret);
  if (ret == XMA_EOS) {
    --m_frames_processing;
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Dequeue() - Returning EOS");
    return ret;
  } else if (ret == XMA_RESEND_AND_RECV) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Dequeue() - Returning resend and recv");
    return XMA_RESEND_AND_RECV;
  }
  data                                 = QueueDataBuffer();
  std::get<QueueDataBuffer>(data).data = std::move(packet);
  --m_frames_processing;
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Dequeue() - Returning success");
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
void InternalEncoderSession::InitParams(XmaEncoderProperties* props) {
  m_mem_ctx = vpi_mem_handle_t(m_sf->CreateMemHandle(), [](VpiMemHandle mem_ctx) { vpi_memory_release(mem_ctx); });
  if (!m_mem_ctx) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Failed to create VPI memory!");
  }

  memset(&m_params, 0, sizeof(VpiEncInitParams));
  m_params.enc_type     = VPE_ENCODER_XAV1;
  m_params.dev_ctx      = m_sf->GetDevHandle();
  m_params.mem_ctx      = m_mem_ctx.get();
  m_params.rate_num     = props->framerate.numerator;
  m_params.rate_denom   = props->framerate.denominator;
  m_params.width        = props->width;
  m_params.height       = props->height;
  m_params.walign       = XMA_FRAME_W_ALIGN;
  m_params.halign       = XMA_FRAME_H_ALIGN;
  m_params.bit_rate     = props->bitrate;
  m_params.qp           = props->qp;
  m_params.gop_size     = props->gop_size;
  m_params.rate_control = static_cast<RateControlMode>(props->rc_mode);
  m_params.enc_level    = props->level;
  m_params.enc_profile  = static_cast<EncProfile>(props->profile);

  if ((props->qp == XMA_ENC_QP_DEFAULT) && (props->bitrate == XMA_ENC_BITRATE_DEFAULT)) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Either QP or bitrate must be set!");
  } else if ((props->qp != XMA_ENC_QP_DEFAULT) && (props->bitrate != XMA_ENC_BITRATE_DEFAULT)) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Both QP and bitrate must not be set!");
  }

  if (props->lookahead_depth <= XMA_ENC_MAX_LOOKAHEAD_DEPTH) {
    m_params.la_depth = props->lookahead_depth;
  } else {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Invalid value for lookahead_depth!");
  }

  if (props->spat_aq_gain <= XMA_ENC_MAX_SPATIAL_AQ_GAIN) {
    m_params.spatial_aq_gain = props->spat_aq_gain;
  } else {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Invalid value for spatial_aq_gain!");
  }

  if (props->temp_aq_gain <= XMA_ENC_MAX_TEMPORAL_AQ_GAIN) {
    m_params.temporal_aq_gain = props->temp_aq_gain;
  } else {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Invalid value for temporal_aq_gain!");
  }

  m_params.preset = const_cast<char*>(GetEnumName(PresetList, sizeof(PresetList) / sizeof(PresetList[0]), props->preset));
  if (m_params.preset == nullptr) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Unsupported preset!");
  }

  int32_t enc_max_val_min_qp, enc_max_val_max_qp = 0;

  if (props->hwencoder_type == XMA_AV1_ENCODER_TYPE) {
    enc_max_val_min_qp = XMA_ENC_AV1_MIN_QP_MAX_VALUE;
    enc_max_val_max_qp = XMA_ENC_AV1_MAX_QP_MAX_VALUE;
  } else {
    enc_max_val_min_qp = XMA_ENC_MIN_QP_MAX_VALUE;
    enc_max_val_max_qp = XMA_ENC_MAX_QP_MAX_VALUE;
  }

  if (props->minQP >= XMA_ENC_QP_MIN_VALUE && props->minQP <= enc_max_val_min_qp) {
    m_params.min_qp = props->minQP;
  } else {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Invalid value for min_qp!");
  }

  if (props->maxQP >= XMA_ENC_QP_MIN_VALUE && props->maxQP <= enc_max_val_max_qp) {
    m_params.max_qp = props->maxQP;
  } else {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Invalid value for max_qp!");
  }

  XmaParameter* param = nullptr;
  VpiHwId       hw_id;
  switch (props->hwencoder_type) {
  case XMA_H264_ENCODER_TYPE:
    hw_id                 = VPE_HW_H264ENC;
    m_params.codec_format = VPE_ENC_H264;
    m_params.profile      = const_cast<char*>(GetEnumName(H264ProfileList, sizeof(H264ProfileList) / sizeof(H264ProfileList[0]), props->profile));
    if (m_params.profile == nullptr) {
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Unknown H.264 profile!");
    }
    switch (props->level) {
    case 0:
      m_params.level = const_cast<char*>("auto");
      break;
    case 10:
    case 11:
    case 12:
    case 13:
    case 20:
    case 21:
    case 22:
    case 30:
    case 31:
    case 32:
    case 40:
    case 41:
    case 42:
    case 50:
    case 51:
    case 52:
      m_params.level = const_cast<char*>(std::to_string(props->level).insert(1, ".").c_str());
      break;
    default:
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Unknown H.264 level!");
    }
    break;
  case XMA_HEVC_ENCODER_TYPE:
    hw_id                 = VPE_HW_HEVCENC;
    m_params.codec_format = VPE_ENC_HEVC;
    m_params.profile      = const_cast<char*>(GetEnumName(HEVCProfileList, sizeof(HEVCProfileList) / sizeof(HEVCProfileList[0]), props->profile));
    if (m_params.profile == nullptr) {
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Unknown HEVC profile!");
    }
    switch (props->level) {
    case 0:
      m_params.level = const_cast<char*>("auto");
      break;
    case 10:
    case 20:
    case 21:
    case 30:
    case 31:
    case 40:
    case 41:
    case 50:
    case 51:
    case 52:
      m_params.level = const_cast<char*>(std::to_string(props->level).insert(1, ".").c_str());
      break;
    default:
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Unknown HEVC level!");
    }
    break;
  case XMA_AV1_ENCODER_TYPE:
    if ((param = GetParameter(props->params, props->param_cnt, XMA_ENC_PARAM_DEVICE_TYPE)) != nullptr) {
      switch (*static_cast<XmaEncoderDeviceType*>(param->value)) {
      case XMA_ENC_DEVICE_TYPE_ANY:
      case XMA_ENC_DEVICE_TYPE_1:
        hw_id                 = VPE_HW_XAV1ENC;
        m_params.codec_format = VPE_ENC_AV1;
        m_params.av1_type     = VPI_ENC_TYPE_1;
        break;
      case XMA_ENC_DEVICE_TYPE_2:
        hw_id                 = VPE_HW_XAV1ENC;
        m_params.codec_format = VPE_ENC_AV1_T2;
        m_params.av1_type     = VPI_ENC_TYPE_2;
        break;
      default:
        THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Invalid device type!");
      }
    } else {
      hw_id                 = VPE_HW_XAV1ENC;
      m_params.codec_format = VPE_ENC_AV1;
      m_params.av1_type     = VPI_ENC_TYPE_1;
    }
    m_params.profile = const_cast<char*>(GetEnumName(AV1ProfileList, sizeof(AV1ProfileList) / sizeof(AV1ProfileList[0]), props->profile));
    if (m_params.profile == nullptr) {
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Unknown AV1 profile!");
    }
    switch (props->level) {
    case 0:
      m_params.level = const_cast<char*>("auto");
      break;
    case 20:
    case 21:
    case 30:
    case 40:
    case 41:
    case 50:
    case 51:
    case 52:
    case 53:
      m_params.level = const_cast<char*>(std::to_string(props->level).insert(1, ".").c_str());
      break;
    default:
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Unknown AV1 level!");
    }
    break;
  default:
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Invalid encoder type!");
  }

  VpiHwCfg* hw_cfg = vpi_device_get_hw_config(m_sf->GetDevHandle(), hw_id);
  if ((props->width < hw_cfg->min_width) || (hw_cfg->max_width < props->width) || (props->height < hw_cfg->min_height) || (hw_cfg->max_height < props->height)) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Unsupported frame size!");
  }

  if (props->format == XMA_VPE_FMT_TYPE) {
    m_params.input_format = props->sw_format;
  } else {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Input format not supported!");
  }

  VpiFmt* pfmt = hw_cfg->sw_formats;
  while (*pfmt != VPI_FORMAT_NONE) {
    if (*pfmt == static_cast<VpiFmt>(props->sw_format)) {
      break;
    }

    ++pfmt;
  }
  if (*pfmt == VPI_FORMAT_NONE) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Input format not supported!");
  }
  vpi_device_free_hw_config(hw_cfg);

  m_params.orig_width  = props->width;
  m_params.orig_height = props->height;

  if ((props->sw_format == XMA_P010LE_FMT_TYPE) || (props->sw_format == XMA_YUV420P10LE_FMT_TYPE) || (props->sw_format == XMA_PACKED10_FMT_TYPE)) {
    m_params.bitdepth = 10;
  } else {
    m_params.bitdepth = 8;
  }

  m_params.flag         = props->flags;
  m_params.ch_index     = 0;
  m_params.enc_index    = 0;
  m_params.xav1_options = NULL;

  param = GetParameter(props->params, props->param_cnt, XMA_ENC_PARAM_WAIT);
  if (param) {
    m_wait = *static_cast<int32_t*>(param->value);
  }

  param = GetParameter(props->params, props->param_cnt, "extradata");
  if (param) {
    m_extra_data = *reinterpret_cast<uint8_t**>(param->value);

    param = GetParameter(props->params, props->param_cnt, "extradata_size");
    if (param) {
      m_extra_data_size = reinterpret_cast<int32_t*>(param->value);
    } else {
      THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Extra data size parameter missing!");
    }
  }

  memset(m_param_str, 0, sizeof(m_param_str));

  char prop[64];

  m_params.slice_id = SetParam(props->params, props->param_cnt, XMA_ENC_PARAM_SLICE, XMA_ENC_SLICE_MIN_VALUE, XMA_ENC_SLICE_1, XMA_ENC_SLICE_DEFAULT);

  m_params.spatial_aq = SetParam(
      props->params, props->param_cnt, XMA_ENC_PARAM_SPATIAL_AQ, XMA_ENC_SPATIAL_AQ_AUTO, XMA_ENC_SPATIAL_AQ_ENABLE, XMA_ENC_SPATIAL_AQ_DEFAULT);

  m_params.temporal_aq = SetParam(
      props->params, props->param_cnt, XMA_ENC_PARAM_TEMPORAL_AQ, XMA_ENC_TEMPORAL_AQ_AUTO, XMA_ENC_TEMPORAL_AQ_ENABLE, XMA_ENC_TEMPORAL_AQ_DEFAULT);

  m_latency_logging = SetParam(
      props->params, props->param_cnt, XMA_ENC_PARAM_LATENCY_LOGGING, XMA_ENC_LATENCY_LOGGING_DISABLE, XMA_ENC_LATENCY_LOGGING_ENABLE, XMA_ENC_LATENCY_LOGGING_DEFAULT);

  m_params.enc_force_idr = SetParam(
      props->params, props->param_cnt, XMA_ENC_PARAM_FORCED_IDR, XMA_ENC_FORCED_IDR_DISABLE, XMA_ENC_FORCED_IDR_ENABLE, XMA_ENC_FORCED_IDR_DEFAULT);

  m_params.crf = SetParam(props->params, props->param_cnt, XMA_ENC_PARAM_CRF, XMA_ENC_PARAM_CRF_DISABLE, XMA_ENC_PARAM_CRF_ENABLE, XMA_ENC_PARAM_CRF_DEFAULT);

  m_params.dynamic_gop = SetParam(
      props->params, props->param_cnt, XMA_ENC_PARAM_DYNAMIC_GOP, XMA_ENC_DYNAMIC_GOP_AUTO, XMA_ENC_DYNAMIC_GOP_ENABLE, XMA_ENC_DYNAMIC_GOP_DEFAULT);

  m_params.tune_metrics = static_cast<TuneMetrics>(
      SetParam(props->params, props->param_cnt, XMA_ENC_PARAM_TUNE_METRICS, XMA_ENC_TUNE_METRICS_VQ, XMA_ENC_TUNE_METRICS_VMAF, XMA_ENC_TUNE_METRICS_DEFAULT));

  m_params.qp_mode = static_cast<QpMode>(
      SetParam(props->params, props->param_cnt, XMA_ENC_PARAM_QP_MODE, XMA_ENC_QP_MODE_AUTO, XMA_ENC_QP_MODE_UNIFORM, XMA_ENC_QP_MODE_DEFAULT));

  if ((param = GetParameter(props->params, props->param_cnt, XMA_ENC_PARAM_BF)) != nullptr) {
    int8_t max_num_bframes;
    if (m_params.av1_type == VPI_ENC_TYPE_1) {
      max_num_bframes = XMA_AV1_AMD_ENC_MAX_BF_VALUE;
    } else {
      max_num_bframes = XMA_ENC_MAX_BF_VALUE;
    }
    m_params.num_b_frames = SetParam(props->params, props->param_cnt, XMA_ENC_PARAM_BF, XMA_ENC_MIN_BF_VALUE, max_num_bframes, XMA_ENC_BF_DEFAULT);
  }

  if ((param = GetParameter(props->params, props->param_cnt, XMA_ENC_PARAM_MAX_BITRATE)) != nullptr) {
    int64_t param_value = *reinterpret_cast<int64_t*>(param->value);
    if (param_value < XMA_ENC_MAX_BITRATE_MIN_VALUE || param_value > XMA_ENC_MAX_BITRATE_MAX_VALUE) {
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Invalid max_bitrate value");
    } else {
      m_params.max_bitrate = *reinterpret_cast<int64_t*>(param->value);
    }
  } else {
    m_params.max_bitrate = XMA_ENC_PARAM_MAX_BITRATE_DEFAULT;
  }

  if ((param = GetParameter(props->params, props->param_cnt, XMA_ENC_PARAM_EXPERT_OPTIONS)) != nullptr) {
    strcpy(prop, (char*) param->value);
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "useMemoryCompression")) != nullptr) {
    sprintf(prop, ",useMemoryCompression=%s", (*(int32_t*) param->value) ? "true" : "false");
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "openGop")) != nullptr) {
    sprintf(prop, ",openGop=%s", (*(int32_t*) param->value) ? "true" : "false");
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "intraRefreshPeriod")) != nullptr) {
    sprintf(prop, ",intraRefreshPeriod=%d", *(int32_t*) param->value);
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "maxM")) != nullptr) {
    sprintf(prop, ",maxM=%d", *(int32_t*) param->value);
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "bll")) != nullptr) {
    sprintf(prop, ",bll=%s", (*(int32_t*) param->value) ? "true" : "false");
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "strictIdr")) != nullptr) {
    sprintf(prop, ",strictIdr=%s", (*(int32_t*) param->value) ? "true" : "false");
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "vbvSize")) != nullptr) {
    sprintf(prop, ",vbvSize=%u", *(uint32_t*) param->value);
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "vbvTarget")) != nullptr) {
    sprintf(prop, ",vbvTarget=%d", *(int32_t*) param->value);
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "needRecon")) != nullptr) {
    sprintf(prop, ",needRecon=%s", (*(int32_t*) param->value) ? "true" : "false");
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "numVTiles")) != nullptr) {
    sprintf(prop, ",numVTiles=%d", *(int32_t*) param->value);
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "numHTiles")) != nullptr) {
    sprintf(prop, ",numHTiles=%d", *(int32_t*) param->value);
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "useBothEncoderInstances")) != nullptr) {
    sprintf(prop, ",useBothEncoderInstances=%s", (*(int32_t*) param->value) ? "true" : "false");
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "tune")) != nullptr) {
    sprintf(prop, ",tune=%d", *(int32_t*) param->value);
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "hdr")) != nullptr) {
    sprintf(prop, ",hdr=%s", (*(int32_t*) param->value) ? "true" : "false");
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "aqEnable")) != nullptr) {
    sprintf(prop, ",aqEnable=%s", (*(int32_t*) param->value) ? "true" : "false");
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "qpMapEnable")) != nullptr) {
    sprintf(prop, ",qpMapEnable=%s", (*(int32_t*) param->value) ? "true" : "false");
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "dynamicGop")) != nullptr) {
    sprintf(prop, ",dynamicGop=%s", (*(int32_t*) param->value) ? "true" : "false");
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "mbtreeEnable")) != nullptr) {
    sprintf(prop, ",mbtreeEnable=%s", (*(int32_t*) param->value) ? "true" : "false");
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "dynamicRoundingEnable")) != nullptr) {
    sprintf(prop, ",dynamicRoundingEnable=%s", (*(int32_t*) param->value) ? "true" : "false");
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "vceVertSearchRange")) != nullptr) {
    sprintf(prop, ",vceVertSearchRange=%d", *(int32_t*) param->value);
    strcat(m_param_str, prop);
  }

  if (m_param_str != NULL) {
    m_params.xav1_options = m_param_str;
  }
}

///////////////////////////////////////////////////////////////////////////////
void InternalEncoderSession::StartEncoder(FramePtr frame) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::StartEncoder()");

  lock_t guard(m_mutex);

  if (m_handle.get()) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::StartEncoder() Encoder already initialized!");
  }

  if (frame) {
    SideDataPtr sd = frame->GetSideData(XMA_FRAME_SIDE_DATA_HDR10_PARAMS);
    if (sd) {
      XmaHdrParams* hdrParams = reinterpret_cast<XmaHdrParams*>(sd->GetHostPtr());
      memcpy(&m_params.mdcv, &hdrParams->mdcv, sizeof(XmaMasterDisColVol));
      memcpy(&m_params.cll, &hdrParams->cll, sizeof(XmaContentLightLevel));
      memcpy(&m_params.atc, &hdrParams->atc, sizeof(XmaATCInfo));
    }
    sd = frame->GetSideData(XMA_FRAME_SIDE_DATA_VUI_PARAMS);
    if (sd) {
      XmaVuiParams* vuiParams = reinterpret_cast<XmaVuiParams*>(sd->GetHostPtr());
      memcpy(&m_params.vui_param, vuiParams, sizeof(XmaVuiParams));
    }
  }

  m_handle = enc_handle_t(vpi_encoder_init(&m_params), [](VpiEncHandle handle) { vpi_encoder_close(handle); });
  if (!m_handle.get()) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::StartEncoder() Failed to initialize encoder!");
  }
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::StartEncoder() enc_handle : %p", m_handle.get());

  if (vpi_encoder_open(m_handle.get()) < 0) {
    m_handle.reset();
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::StartEncoder() Unable to open encoder for processing!");
  }

  VpiEncContext* enc_ctx      = static_cast<VpiEncContext*>(m_handle.get());
  Xav1Context*   enc_xav1_ctx = static_cast<Xav1Context*>(enc_ctx->vpi_encoder_handle);
  m_max_frames_processing     = enc_xav1_ctx->max_frames_delay;
  if (m_params.codec_format == VPE_ENC_AV1) {
    m_max_frames_processing += XAV1_EXTRA_FRAMES;
  }
  if (m_params.la_depth == 0) {
    m_max_frames_processing = 1;
  }

  if (m_extra_data && m_extra_data_size) {
    uint8_t* extradata_buffer = reinterpret_cast<uint8_t*>(vpi_encoder_get_extradata(m_handle.get(), m_extra_data_size));
    memcpy(m_extra_data, extradata_buffer, *m_extra_data_size);
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::StartEncoder() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalEncoderSession::GetPacket(IntDataBufferPtr& buf) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::GetPacket()");
  VpiPkt pkt;
  memset(&pkt, 0, sizeof(VpiPkt));
  for (int32_t retry_count = XMA_ENCODER_MAX_RETRY_COUNT; retry_count >= 0; --retry_count) {
    pkt.buffer = static_cast<uint8_t*>(buf->GetBuffer());
    pkt.size   = buf->GetSize();

    if (vpi_encoder_get_packet(m_handle.get(), &pkt) != 0) {
      THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::GetPacket() Failed to get packet!");
    }

    if (pkt.size == 0) {
      if (m_end_found) {
        if (pkt.is_eof) {
          m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::GetPacket() - Returning end of file");
          return XMA_EOS;
        }
        if (retry_count > 0) {
          if (vpi_encoder_put_frame(m_handle.get(), nullptr) < 0) {
            THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::GetPacket() Failed to send NULL to encoder!");
          }
          usleep(static_cast<uint32_t>(1000000L * m_params.rate_denom / (m_params.rate_num * XMA_ENCODER_RETRY_DELAY)));
          continue;
        } else {
          THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::GetPacket() Recieve failed during flush!");
        }
      } else if ((m_wait) && (m_frames_processing >= m_max_frames_processing)) {
        if (retry_count > 0) {
          usleep(static_cast<uint32_t>(1000000L * m_params.rate_denom / (m_params.rate_num * XMA_ENCODER_RETRY_DELAY)));
          continue;
        } else {
          THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::GetXAV1Packet() Timeout waiting for encoder to get packet!");
        }
      } else {
        m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::GetPacket() - Returning resend and recv");
        return XMA_RESEND_AND_RECV;
      }
    } else {
      break;
    }
  }

  m_frame_recv++;
  if (m_latency_logging) {
    clock_gettime(CLOCK_REALTIME, &m_latency);
    m_time_taken = (m_latency.tv_sec * 1e3) + (m_latency.tv_nsec / 1e6);
    m_log->LogMsg(
        XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::GetPacket() : %p : xma_enc_frame_recv : %lld : %lld\n", m_handle.get(), m_frame_recv, m_time_taken);
  }

  buf->SetSize(pkt.size);
  buf->SetPts(pkt.pts);

  if (m_params.codec_format != VPE_ENC_AV1_T2) {
    m_log->LogMsg(XMA_INFO_LOG, XMA_ENCODER_SESSION, "Encoder PSNR Y=%f U=%f V=%f", pkt.psnr[0], pkt.psnr[1], pkt.psnr[2]);
    m_log->LogMsg(XMA_INFO_LOG, XMA_ENCODER_SESSION, "Encoder SSIM Y=%f U=%f V=%f", pkt.ssim[0], pkt.ssim[1], pkt.ssim[2]);
  }

  PrivateVpiFrmInfo* info = static_cast<PrivateVpiFrmInfo*>(pkt.opaque);
  if (info && info->IsValid()) {
    delete static_cast<EncoderFrame*>(info->GetFreeData());
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::GetPacket() - Returning success");
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
IntDataBufferPtr InternalEncoderSession::PullPacket(int32_t* ret) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::PullPacket()");
  int size = 0;
  vpi_encoder_get_packet_size(m_handle.get(), &size);

  IntDataBufferPtr buf = std::make_shared<InternalDataBuffer>(m_sf, m_log, size, false);

  *ret = GetPacket(buf);
  if (*ret != XMA_SUCCESS) {
    return nullptr;
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::PullPacket() - Returning success");
  return buf;
}

} // namespace xma
