/*
 * Copyright (C) 2021, Xilinx Inc - All rights reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

#include <memory>
#include <cstring>
#ifdef __GNUC__
#include <fcntl.h>
#endif
#include "InternalScalerSession.h"
#include "InternalFrame.h"
#include "xmaapiscaler.h"

namespace xma {

#define XMA_SCALER_SESSION "ScalerSession"

#ifndef MIN
#define MIN(a, b) (((a) < (b)) ? (a) : (b))
#endif

///////////////////////////////////////////////////////////////////////////////
InternalScalerSession::InternalScalerSession(IntSessionFactoryPtr sf, XmaScalerProperties* props) :
    FilterThread<ScalerDynamicParams>(sf, props->params, props->param_cnt, 1, 1) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_SESSION, "InternalScalerSession::InternalScalerSession()");

  m_mem_ctx = vpi_mem_handle_t(m_sf->CreateMemHandle(), [](VpiMemHandle mem_ctx) { vpi_memory_release(mem_ctx); });
  if (!m_mem_ctx) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SCALER_SESSION, "InternalScalerSession::InternalScalerSession() Failed to create VPI memory!");
  }

  if (props->input.format != XMA_VPE_FMT_TYPE) {
    Close();
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SCALER_SESSION, "InternalScalerSession::InternalScalerSession() Input must be on device!");
  }
  if (((props->input.width % 2) != 0) || ((props->input.height % 2) != 0)) {
    Close();
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SCALER_SESSION, "InternalScalerSession::InternalScalerSession() Input size(%dx%d) must be multiple of 2!",
        props->input.width, props->input.height);
  }

  m_outputs   = props->num_outputs;
  m_left      = 0;
  m_top       = 0;
  m_width     = props->input.width;
  m_height    = props->input.height;
  m_framerate = props->input.framerate;

  XmaParameter* param = GetParameter(props->params, props->param_cnt, XMA_SCALER_PARAM_DYNAMIC_CROPPING);
  if (param) {
    m_dynamic_cropping = *static_cast<int32_t*>(param->value);
  }
  if (m_dynamic_cropping) {
    param = GetParameter(props->params, props->param_cnt, XMA_SCALER_PARAM_LEFT_CROPPING);
    if (param) {
      m_dynamic_left = static_cast<int32_t*>(param->value);
    }
    param = GetParameter(props->params, props->param_cnt, XMA_SCALER_PARAM_TOP_CROPPING);
    if (param) {
      m_dynamic_top = static_cast<int32_t*>(param->value);
    }
    param = GetParameter(props->params, props->param_cnt, XMA_SCALER_PARAM_WIDTH_CROPPING);
    if (param) {
      m_dynamic_width = static_cast<int32_t*>(param->value);
    }
    param = GetParameter(props->params, props->param_cnt, XMA_SCALER_PARAM_HEIGHT_CROPPING);
    if (param) {
      m_dynamic_height = static_cast<int32_t*>(param->value);
    }
  } else {
    param = GetParameter(props->params, props->param_cnt, XMA_SCALER_PARAM_LEFT_CROPPING);
    if (param) {
      m_left = *static_cast<int32_t*>(param->value);
    }
    param = GetParameter(props->params, props->param_cnt, XMA_SCALER_PARAM_TOP_CROPPING);
    if (param) {
      m_top = *static_cast<int32_t*>(param->value);
    }
    param = GetParameter(props->params, props->param_cnt, XMA_SCALER_PARAM_WIDTH_CROPPING);
    if (param) {
      m_width = MIN(*static_cast<int32_t*>(param->value), props->input.width - m_left);
    }
    param = GetParameter(props->params, props->param_cnt, XMA_SCALER_PARAM_HEIGHT_CROPPING);
    if (param) {
      m_height = MIN(*static_cast<int32_t*>(param->value), props->input.height - m_top);
    }
  }
  if ((param = GetParameter(props->params, props->param_cnt, XMA_SCALER_PARAM_LATENCY_LOGGING)) != nullptr) {
    if (*(int32_t*) param->value < 0 || *(int32_t*) param->value > 1) {
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SCALER_SESSION, "InternalScalerSession::InternalScalerSession() Invalid value for latency_logging");
    } else {
      m_latency_logging = *(int32_t*) param->value;
    }
  } else {
    m_latency_logging = XMA_SCALER_LATENCY_LOGGING_DEFAULT;
  }

  memset(&m_params, 0, sizeof(VpiProcInitParams));
  m_params.processor_type  = VPE_PROCESSOR_XABR;
  m_params.dev_ctx         = m_sf->GetDevHandle();
  m_params.mem_ctx         = m_mem_ctx.get();
  m_params.output_num      = props->num_outputs;
  m_params.resize_str      = m_resize_str;
  m_params.width           = props->input.width;
  m_params.height          = props->input.height;
  m_params.walign          = XMA_FRAME_W_ALIGN;
  m_params.halign          = XMA_FRAME_H_ALIGN;
  m_params.flag            = props->input.flags;
  m_params.channel_choose  = 0;
  m_params.input_format    = InternalFrameProperties::GetPixFmtName(props->input.format);
  m_params.input_sw_format = InternalFrameProperties::GetPixFmtName(props->input.sw_format);

  param = GetParameter(props->params, props->param_cnt, XMA_SCALER_PARAM_MIX_RATE);
  if (param) {
    void* mixrate           = static_cast<void*>(param->value);
    m_mixrate_parent_scaler = GetMixrateScaler(mixrate, props);
    if (!m_mixrate_parent_scaler) {
      m_log->LogMsg(XMA_ERROR_LOG, XMA_SCALER_SESSION, "InternalScalerSession::InternalScalerSession() mix_rate does not specify a valid scaler session!");
      Close();
      throw InvalidArgumentException("InternalScalerSession::InternalScalerSession() mix_rate does not specify a valid scaler session!");
    }
    m_mixrate_scaler = m_mixrate_parent_scaler.get();
    m_mixrate_parent_scaler->SetMixrateChild(this);
  }

  int32_t output;
  for (output = 0; output < m_outputs; ++output) {
    if (((props->output[output].width % 2) != 0) || ((props->output[output].height % 2) != 0)) {
      Close();
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SCALER_SESSION,
          "InternalScalerSession::InternalScalerSession() Output size(%dx%d) must be multiple of 2!", props->output[output].width, props->output[output].height);
    }

    m_props.push_back(std::make_shared<InternalFrameProperties>(m_sf, m_log, props->output[output].format, props->output[output].sw_format, props->output[output].width,
        props->output[output].height, props->output[output].flags));
    char size[128];
    char out_fmt_str[64];
    out_fmt_str[0] = '\0';
    strcat(out_fmt_str, InternalFrameProperties::GetPixFmtName(m_props[output]->GetSwFormat()));
    uint32_t flags = m_props[output]->GetFlags() & (XMA_FRAME_PROPERTY_FLAG_TILE_4x4 | XMA_FRAME_PROPERTY_FLAG_COMPRESS);
    if (flags == (XMA_FRAME_PROPERTY_FLAG_TILE_4x4 | XMA_FRAME_PROPERTY_FLAG_COMPRESS)) {
      strcat(out_fmt_str, "-ma");
    } else if (flags == XMA_FRAME_PROPERTY_FLAG_TILE_4x4) {
      strcat(out_fmt_str, "-tile");
    } else if (flags != 0) {
      Close();
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SCALER_SESSION, "InternalScalerSession::InternalScalerSession() Unsupported output flags!");
    }
    sprintf(size, "(%dx%d|%s)", m_props[output]->GetWidth(), m_props[output]->GetHeight(), out_fmt_str);
    strcat(m_resize_str, size);

    if (props->output[output].format != XMA_VPE_FMT_TYPE) {
      Close();
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SCALER_SESSION, "InternalScalerSession::InternalScalerSession() All outputs must be on device!");
    }
    if ((props->output[output].framerate.numerator != m_framerate.numerator) || (props->output[output].framerate.denominator != m_framerate.denominator)) {
      Close();
      THROW_INVALID_ARGUMENT_EXCEPTION(
          m_log, XMA_ERROR_LOG, XMA_SCALER_SESSION, "InternalScalerSession::InternalScalerSession() All outputs must be same framerate as source!");
    }
  }
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_SESSION, "InternalScalerSession::InternalScalerSession() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
InternalScalerSession::~InternalScalerSession() {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_SESSION, "InternalScalerSession::~InternalScalerSession()");

  Close();

  m_handle.reset();
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalScalerSession::Enqueue(queue_t& data) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_SESSION, "InternalScalerSession::Enqueue()");

  if ((!std::get<QueueFrame>(data).frame) && (m_mixrate_scaler)) {
    m_mixrate_scaler->SendMixrateEOS();
  }

  ScalerDynamicParams dynamic_params;
  if (m_dynamic_cropping) {
    dynamic_params.left   = (m_dynamic_left == nullptr) ? m_left : (*m_dynamic_left);
    dynamic_params.top    = (m_dynamic_top == nullptr) ? m_top : (*m_dynamic_top);
    dynamic_params.width  = (m_dynamic_width == nullptr) ? m_width : (*m_dynamic_width);
    dynamic_params.height = (m_dynamic_height == nullptr) ? m_height : (*m_dynamic_height);
  }
  int32_t ret = FilterThread::Push(data, &dynamic_params);

  switch (ret) {
  case XMA_TRY_AGAIN:
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_SESSION, "InternalScalerSession::Enqueue() - Returning try again");
    break;
  case XMA_SUCCESS:
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_SESSION, "InternalScalerSession::Enqueue() - Returning success");
    break;
  default:
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_SESSION, "InternalScalerSession::Enqueue() - Returning");
    break;
  }
  return ret;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalScalerSession::Dequeue(queue_t& data) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_SESSION, "InternalScalerSession::Dequeue()");

  int32_t ret = FilterThread::Pull(data);

  if ((ret == XMA_EOS) && (m_mixrate_scaler)) {
    m_mixrate_scaler->RecvMixrateEOS();
  }

  char error_msg[1024];
  sprintf(error_msg, "InternalScalerSession::Dequeue() - returning %d", ret);
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_SESSION, error_msg);

  return ret;
}

///////////////////////////////////////////////////////////////////////////////
void InternalScalerSession::SetMixrateChild(InternalScalerSession* child) {
  m_mixrate_scaler = child;
}

///////////////////////////////////////////////////////////////////////////////
void InternalScalerSession::SendMixrateEOS() {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_SESSION, "InternalScalerSession::SendMixrateEOS()");

  queue_t             data = QueueFrame();
  ScalerDynamicParams dynamic_params;
  memset(&dynamic_params, 0, sizeof(dynamic_params));

  FilterThread::Push(data, &dynamic_params);
}

///////////////////////////////////////////////////////////////////////////////
void InternalScalerSession::RecvMixrateEOS() {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_SESSION, "InternalScalerSession::RecvMixrateEOS()");

  queue_t data = QueueFrameList();
  int32_t ret  = FilterThread::Pull(data);
  if (ret != XMA_EOS) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SCALER_SESSION, "InternalScalerSession::RecvMixrateEOS() Did not recieve EOS!");
  }
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalScalerSession::Process(queue_t& in, queue_t& out, ScalerDynamicParams* dynamic_params) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_SESSION, "InternalScalerSession::Process()");
  FramePtr frame = std::move(std::get<QueueFrame>(in).frame);

  if (!frame) {
    out = QueueFrameList();
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_SESSION, "InternalScalerSession::Process() - Returning EOS");
    return XMA_EOS;
  }

  VpiFrm* pic = nullptr;
  if ((vpi_frame_alloc(m_sf->GetMemHandle(), &pic) < 0) || !pic) {
    THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SCALER_SESSION, "InternalScalerSession::Process() Out of memory!");
  }
  InitVpiFrm(pic);
  using vpi_frm_t             = std::unique_ptr<VpiFrm, std::function<void(VpiFrm*)>>;
  IntSessionFactoryPtr sf     = m_sf;
  vpi_frm_t            in_pic = vpi_frm_t(pic, [sf](VpiFrm* frm) {
    vpi_frame_unref(frm);
    if (vpi_buffer_get_ref_count(static_cast<VpiBuffer*>(frm->vpi_picinfo_buf)) == 0) {
      vpi_frame_free(sf->GetMemHandle(), frm);
    }
  });

  int32_t    outputs  = m_outputs;
  vpi_frms_t out_pics = vpi_frms_t(new VpiFrm*[MAX_SCALER_OUTPUTS], [outputs, sf](VpiFrm** frms) {
    int32_t frm;
    for (frm = 0; frm < outputs; ++frm) {
      if (frms[frm]) {
        vpi_frame_unref(frms[frm]);
        if (vpi_buffer_get_ref_count(static_cast<VpiBuffer*>(frms[frm]->vpi_picinfo_buf)) == 0) {
          vpi_frame_free(sf->GetMemHandle(), frms[frm]);
        }
      }
    }
    delete[] frms;
  });
  int32_t    output;
  for (output = 0; output < m_outputs; ++output) {
    if ((vpi_frame_alloc(m_sf->GetMemHandle(), &out_pics.get()[output]) < 0) || !out_pics.get()[output]) {
      THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SCALER_SESSION, "InternalScalerSession::Enque() Out of memory!");
    }
    InitVpiFrm(out_pics.get()[output]);
  }

  int32_t ret;
  if ((ret = AddFrameToVpiFrm(m_sf, m_log, frame, in_pic.get())) != XMA_SUCCESS) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SCALER_SESSION, "InternalScalerSession::Process() Failed to convert XmaFrame to VpiFrm!");
  }

  lock_t guard; // only lock if threads are being used, otherwise FilterThread::Pull() will lock
  if (m_num_threads > 0) {
    lock_t temp_guard(m_mutex);
    guard.swap(temp_guard);
  }

  if (m_dynamic_cropping) {
    int32_t left;
    int32_t top;
    int32_t width;
    int32_t height;
    if (m_dynamic_left) {
      left = dynamic_params->left;
      if ((left < 0) || (left >= m_width)) {
        THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SCALER_SESSION, "InternalScalerSession::Process() Invalid left position(%d)", left);
      }
    } else {
      left = 0;
    }
    if (m_dynamic_top) {
      top = dynamic_params->top;
      if ((top < 0) || (top >= m_height)) {
        THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SCALER_SESSION, "InternalScalerSession::Process() Invalid top position(%d)", top);
      }
    } else {
      top = 0;
    }
    if (m_dynamic_width) {
      width = dynamic_params->width;
      if (width < 0) {
        THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SCALER_SESSION, "InternalScalerSession::Process() Invalid width(%d)", width);
      }
      if ((left + width) > m_width) {
        THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SCALER_SESSION,
            "InternalScalerSession::Process() Invalid width! Left(%d) + width(%d) exceeds frame width(%d)!", left, width, m_width);
      }
    } else {
      width = m_width - left;
    }
    if (m_dynamic_height) {
      height = dynamic_params->height;
      if (height < 0) {
        THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SCALER_SESSION, "InternalScalerSession::Process() Invalid height(%d)", height);
      }
      if ((top + height) > m_height) {
        THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SCALER_SESSION,
            "InternalScalerSession::Process() Invalid height! Top(%d) + height(%d) exceeds frame height(%d)!", left, height, m_height);
      }
    } else {
      height = m_height - top;
    }

    if (((left % 2) != 0) || ((top % 2) != 0)) {
      THROW_INVALID_ARGUMENT_EXCEPTION(
          m_log, XMA_ERROR_LOG, XMA_SCALER_SESSION, "InternalScalerSession::Process() Crop position(%dx%d) must be multiple of 2!", left, top);
    }
    if (((width % 2) != 0) || ((height % 2) != 0)) {
      THROW_INVALID_ARGUMENT_EXCEPTION(
          m_log, XMA_ERROR_LOG, XMA_SCALER_SESSION, "InternalScalerSession::Process() Crop size(%dx%d) must be multiple of 2!", width, height);
    }
    InitScaler(left, top, MIN(width, m_width - left), MIN(height, m_height - top));
  } else {
    InitScaler(m_left, m_top, m_width, m_height);
  }

  m_frame_sent++;
  if (m_latency_logging) {
    clock_gettime(CLOCK_REALTIME, &m_latency);
    m_time_taken = (m_latency.tv_sec * 1e3) + (m_latency.tv_nsec / 1e6);
    m_log->LogMsg(
        XMA_DEBUG_LOG, XMA_SCALER_SESSION, "InternalScalerSession::Process() : %p : xma_scaler_frame_sent : %lld : %lld\n", m_handle.get(), m_frame_sent, m_time_taken);
  }

  if (vpi_processor_process_frame_list(m_handle.get(), in_pic.release(), out_pics.get()) < 0) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SCALER_SESSION, "InternalScalerSession::Process() Failed to process frame in scaler!");
  }

  m_frame_recv++;
  if (m_latency_logging) {
    clock_gettime(CLOCK_REALTIME, &m_latency);
    m_time_taken = (m_latency.tv_sec * 1e3) + (m_latency.tv_nsec / 1e6);
    m_log->LogMsg(
        XMA_DEBUG_LOG, XMA_SCALER_SESSION, "InternalScalerSession::Process() : %p : xma_scaler_frame_recv : %lld : %lld\n", m_handle.get(), m_frame_recv, m_time_taken);
  }

  out                                      = QueueFrameList();
  std::get<QueueFrameList>(out).frame_list = std::vector<FramePtr>(m_outputs);

  for (output = 0; output < m_outputs; ++output) {
    IntFramePtr out_frame = std::make_shared<InternalFrame>(m_sf, m_log, out_pics.get()[output], m_framerate, m_props[output]->GetFormat(),
        m_props[output]->GetSwFormat(), m_props[output]->GetWidth(), m_props[output]->GetHeight());
    if (!out_frame) {
      std::get<QueueFrameList>(out).frame_list.clear();
      THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SCALER_SESSION, "InternalScalerSession::Process() Out of memory!");
    }

    // To Do: Remove this once the scaler passes this side data
    SideDataPtr sd = frame->GetSideData(XMA_FRAME_SIDE_DATA_HDR10_PARAMS);
    if (sd && (!out_frame->GetSideData(XMA_FRAME_SIDE_DATA_HDR10_PARAMS))) {
      out_frame->AddSideData(sd);
    }

    std::get<QueueFrameList>(out).frame_list[output] = out_frame;
  }

  delete out_pics.release();
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_SESSION, "InternalScalerSession::Process() - Returning success");
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
void InternalScalerSession::InitScaler(int32_t left, int32_t top, int32_t width, int32_t height) {
  if ((left == m_old_left) && (top == m_old_top) && (width == m_old_width) && (height == m_old_height)) {
    return;
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_SESSION, "InternalScalerSession::InitScaler(top x left=%dx%d, width x height=%dx%d)", left, top, width, height);
  m_handle.reset();

  if (((m_width % 2) != 0) || ((m_height % 2) != 0)) {
    THROW_INVALID_ARGUMENT_EXCEPTION(
        m_log, XMA_ERROR_LOG, XMA_SCALER_SESSION, "InternalScalerSession::InitScaler() Crop size(%dx%d) must be multiple of 2!", width, height);
  }

  VpiHwCfg* hw_cfg = vpi_device_get_hw_config(m_sf->GetDevHandle(), VPE_HW_XABR);
  if ((m_width < hw_cfg->min_width) || (hw_cfg->max_width < m_width) || (m_height < hw_cfg->min_height) || (hw_cfg->max_height < m_height)) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SCALER_SESSION, "InternalScalerSession::InitScaler() Unsupported frame size!");
  }
  vpi_device_free_hw_config(hw_cfg);

  if (m_mixrate_parent_scaler) {
    m_params.mixrate_ctx = static_cast<VpiProcHandle*>(m_mixrate_parent_scaler->m_handle.get());
  }

  VpiProcHandle new_handle = nullptr;
  VpiRet        ret        = vpi_processor_init(&m_params, &new_handle);
  if (ret < VPI_SUCCESS) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SCALER_SESSION, "InternalScalerSession::InitScaler() Failed to initialize scaler: %s!", vpi_error_str(ret));
  }
  m_handle = scale_handle_t(new_handle, [](void* handle) { vpi_processor_close(handle); });

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_SESSION, "InternalScalerSession::InternalScalerSession() scaler_handle : %p", m_handle.get());

  char crop_str[64];
  snprintf(crop_str, sizeof(crop_str), "%d|%d|%d|%d", width, height, left, top);
  vpi_processor_set_param(m_handle.get(), const_cast<char*>("crop_str"), crop_str);

  if (vpi_processor_open(m_handle.get()) < 0) {
    m_handle.reset();
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SCALER_SESSION, "InternalScalerSession::InitScaler() Unable to open scaler for processing");
  }

  m_old_left   = left;
  m_old_top    = top;
  m_old_width  = width;
  m_old_height = height;
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_SESSION, "InternalScalerSession::InitScaler() - Returning");
}

#ifdef __GNUC__
///////////////////////////////////////////////////////////////////////////////
// If this was compiled using GNU, the easiest way to check if the
// user passed a bad pointer is to try writing the pointer to a file,
// the write will fail if it is bad
std::shared_ptr<InternalScalerSession> InternalScalerSession::GetMixrateScaler(void* mixrate, XmaScalerProperties* props) {
  if (props->handle) {
    XmaScalerSession* c_scaler = static_cast<XmaScalerSession*>(mixrate);
    if (c_scaler) {
      int nullfd = open("/dev/null", O_WRONLY);
      if (write(nullfd, c_scaler, sizeof(XmaScalerSession)) < 0) {
        close(nullfd);
        m_log->LogMsg(XMA_ERROR_LOG, XMA_SCALER_SESSION, "InternalScalerSession::InternalScalerSession() mix_rate does not specify a valid scaler session!");
        Close();
        throw InvalidArgumentException("InternalScalerSession::InternalScalerSession() mix_rate does not specify a valid scaler session!");
      }
      close(nullfd);
      return std::dynamic_pointer_cast<InternalScalerSession>(c_scaler->m_session);
    }
  } else {
    ScalerSession* cpp_scaler = static_cast<ScalerSession*>(mixrate);
    if (cpp_scaler) {
      int nullfd = open("/dev/null", O_WRONLY);
      if (write(nullfd, cpp_scaler, sizeof(InternalScalerSession)) < 0) {
        close(nullfd);
        m_log->LogMsg(XMA_ERROR_LOG, XMA_SCALER_SESSION, "InternalScalerSession::InternalScalerSession() mix_rate does not specify a valid scaler session!");
        Close();
        throw InvalidArgumentException("InternalScalerSession::InternalScalerSession() mix_rate does not specify a valid scaler session!");
      }
      close(nullfd);
      return std::dynamic_pointer_cast<InternalScalerSession>(cpp_scaler->shared_from_this());
    }
  }
  return nullptr;
}

#elif defined(_MSC_VER)
///////////////////////////////////////////////////////////////////////////////
// If this was compiled using MSVC, the easiest way to check if the user
// passed a bad pointer as mixrate is to use a __try/__except block
std::shared_ptr<InternalScalerSession> InternalScalerSession::GetMixrateScaler(void* mixrate, XmaScalerProperties* props) {
  __try {
    if (props->handle) {
      XmaScalerSession* c_scaler = static_cast<XmaScalerSession*>(mixrate);
      if (c_scaler) {
        return std::dynamic_pointer_cast<InternalScalerSession>(c_scaler->m_session);
      }
    } else {
      ScalerSession* cpp_scaler = static_cast<ScalerSession*>(mixrate);
      if (cpp_scaler) {
        return std::dynamic_pointer_cast<InternalScalerSession>(cpp_scaler->shared_from_this());
      }
    }
  } __except (EXCEPTION_EXECUTE_HANDLER) {
  }
  return nullptr;
}

#else
///////////////////////////////////////////////////////////////////////////////
// If this was compiled using some other compiler, no checking for the validity
// of mixrate is done and it is assumed it is valid
std::shared_ptr<InternalScalerSession> InternalScalerSession::GetMixrateScaler(void* mixrate, XmaScalerProperties* props) {
  if (props->handle) {
    XmaScalerSession* c_scaler = static_cast<XmaScalerSession*>(mixrate);
    if (c_scaler) {
      return std::dynamic_pointer_cast<InternalScalerSession>(c_scaler->m_session);
    }
  } else {
    ScalerSession* cpp_scaler = static_cast<ScalerSession*>(mixrate);
    if (cpp_scaler) {
      return std::dynamic_pointer_cast<InternalScalerSession>(cpp_scaler->shared_from_this());
    }
  }
  return nullptr;
}
#endif

} // namespace xma
